#!/usr/bin/env python
# -*- coding: utf-8 -*-

from durable.lang import *

with ruleset('animal'):
    # will be triggered by 'Kermit eats flies'
    @when_all((m.predicate == 'eats') & (m.object == 'flies'))
    def frog(c):
        c.assert_fact({ 'subject': c.m.subject, 'predicate': 'is', 'object': 'frog' })

    @when_all((m.predicate == 'eats') & (m.object == 'worms'))
    def bird(c):
        c.assert_fact({ 'subject': c.m.subject, 'predicate': 'is', 'object': 'bird' })

    # will be chained after asserting 'Kermit is frog'
    @when_all((m.predicate == 'is') & (m.object == 'frog'))
    def green(c):
        c.assert_fact({ 'subject': c.m.subject, 'predicate': 'is', 'object': 'green' })

    @when_all((m.predicate == 'is') & (m.object == 'bird'))
    def black(c):
        c.assert_fact({ 'subject': c.m.subject, 'predicate': 'is', 'object': 'black' })

    @when_all(+m.subject)
    def output(c):
        print('Fact: {0} {1} {2}'.format(c.m.subject, c.m.predicate, c.m.object))

assert_fact('animal', { 'subject': 'Kermit', 'predicate': 'eats', 'object': 'flies' })




from typing import NamedTuple


class String(NamedTuple):
    value: str


class Identifier(NamedTuple):
    value: str


class OperatorExpression(NamedTuple):
    left: str
    operator: str
    right: str




from typing import NamedTuple, Union, List


class EventSource(NamedTuple):
    name: str
    source_name: str
    source_args: dict
    transform: Union[str, None]


class Action(NamedTuple):
    module: str
    module_args: dict


class Condition(NamedTuple):
    value: str


class Rule(NamedTuple):
    name: str
    condition: Condition
    action: Action


class RuleSet(NamedTuple):
    name: str
    hosts: Union[str, List[str]]
    sources: List[EventSource]
    rules: List[Rule]





"""
Usage:
    ftl-events [options] <rules.yml> <vars.yml> <inventory.yml>

Options:
    -h, --help            Show this page
    --redis_host_name=<h> Redis host name
    --redis_port=<p>      Redis port
    --debug               Show debug logging
    --verbose             Show verbose logging
"""
from docopt import docopt
import os
import logging
import sys
import yaml
import ftl_events.rules_parser as rules_parser
import ftl_events.rule_generator as rule_generator
from ftl_events.durability import provide_durability
import multiprocessing as mp
import runpy
import jinja2
import asyncio
import durable.lang
from urllib.parse import urlparse
from faster_than_light import run_module, load_inventory

logger = logging.getLogger('cli')


def load_vars(parsed_args):
    with open(parsed_args['<vars.yml>']) as f:
        return yaml.safe_load(f.read())


def load_rules(parsed_args):
    with open(parsed_args['<rules.yml>']) as f:
        return rules_parser.parse_rule_sets(yaml.safe_load(f.read()))


def substitute_variables(value, context):
    return jinja2.Template(value, undefined=jinja2.StrictUndefined).render(context)


def start_sources(sources, variables, queue):

    for source in sources:
        module = runpy.run_path(os.path.join('sources', source.source_name + '.py'))
        args = {k: substitute_variables(v, variables) for k, v in source.source_args.items()}
        module.get('main')(queue, args)


def run_ruleset(ruleset, variables, inventory, queue, redis_host_name=None, redis_port=None):

    if redis_host_name and redis_port:
        provide_durability(durable.lang.get_host(), redis_host_name, redis_port)

    print([ruleset])
    durable_ruleset = rule_generator.generate_rulesets([ruleset], variables)
    print([x.define() for x in durable_ruleset])

    while True:
        data = queue.get()
        print(data)
        if not data:
            continue
        print(data)
        print(ruleset.name)
        try:
            durable.lang.post(ruleset.name, data)
        except durable.engine.MessageNotHandledException:
            print(f'MessageNotHandledException: {data}')

def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parsed_args = docopt(__doc__, args)
    if parsed_args['--debug']:
        logging.basicConfig(level=logging.DEBUG)
    elif parsed_args['--verbose']:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    variables = load_vars(parsed_args)
    rulesets = load_rules(parsed_args)
    inventory = load_inventory(parsed_args['<inventory.yml>'])

    print(variables)
    print(rulesets)

    tasks = []

    for ruleset in rulesets:
        sources = ruleset.sources
        queue = mp.Queue()

        tasks.append(mp.Process(target=start_sources, args=(sources, variables, queue)))
        tasks.append(mp.Process(target=run_ruleset, args=(ruleset, variables, inventory, queue, parsed_args['--redis_host_name'], parsed_args['--redis_port'])))

    for task in tasks:
        task.start()

    for task in tasks:
        task.join()

    return 0

def entry_point():
    main(sys.argv[1:])





from typing import NamedTuple
import durable.lang
from durable.lang import *
from pyparsing import pyparsing_common, infix_notation, OpAssoc, one_of, ParserElement, QuotedString
ParserElement.enable_packrat()
from ftl_events.condition_parser import condition as condition_parser
from ftl_events.rule_generator import visit_condition


def call(args, **kwargs):
    print('args', args)
    for i in args:
        print(type(i))
        for j in i:
            print(type(j))
    return args


class String(NamedTuple):
    value: str


class Identifier(NamedTuple):
    value: str


class OperatorExpression(NamedTuple):
    left: str
    operator: str
    right: str

integer = pyparsing_common.signed_integer
varname = pyparsing_common.identifier.copy().add_parse_action(lambda toks: Identifier(toks[0]))


string1 = QuotedString("'").copy().add_parse_action(lambda toks: String(toks[0]))
string2 = QuotedString('"').copy().add_parse_action(lambda toks: String(toks[0]))

arith_expr = infix_notation(integer | varname | string1 | string2,
                            [
                                ('!', 1, OpAssoc.RIGHT, call),
                                (one_of('* /'), 2, OpAssoc.LEFT),
                                (one_of('+ -'), 2, OpAssoc.LEFT),
                                (one_of('< >'), 2, OpAssoc.LEFT),
                                ('!=', 2, OpAssoc.LEFT, lambda toks: OperatorExpression(*toks[0])),
                                ('==', 2, OpAssoc.LEFT),
                                ('>=', 2, OpAssoc.LEFT),
                                ('<=', 2, OpAssoc.LEFT),
                            ])

def test_infix():
    arith_expr.run_tests('''
        5+3*6
        (5+3)*6
        -2--11
        a!=4
        a==4
        a>4
        a>=4
        a<=4
        a<4
        a != 'hey'
        !True
        False
        true
        false
        ''', full_dump=False)


def visit(parsed_condition, condition):
    if isinstance(parsed_condition, Identifier):
        return condition.__getattr__(parsed_condition.value)
    if isinstance(parsed_condition, String):
        return parsed_condition.value
    if isinstance(parsed_condition, OperatorExpression):
        if parsed_condition.operator == "!=":
            return visit(parsed_condition.left, condition).__ne__(visit(parsed_condition.right, condition))


def test_m():
    assert m
    assert m.x
    assert m.x.define() == {'$m': 'x'}
    assert m.x > m.y
    assert (m.x > m.y).define() == {'$gt': {'x': {'$m': 'y'}}}
    assert m.x == m.y
    assert (m.x == m.y).define() == {'x': {'$m': 'y'}}
    assert m.x < m.y
    assert (m.x < m.y).define() == {'$lt': {'x': {'$m': 'y'}}}
    arith_expr.run_tests('''
                         x
                         x > y
                         x == y
                         x < y
                         text != ""
    ''', full_dump=False)

    result = arith_expr.parseString('text')[0]
    print(result)
    print(visit(result, m).define())

    result = arith_expr.parseString('""')[0]
    print(result)
    print(visit(result, m))

    result = arith_expr.parseString('text != ""')[0]
    print(result)
    print(visit(result, m).define())
    print((m.text != "").define())
    assert visit(result, m).define() == (m.text != "").define()

    result = arith_expr.parseString('x != y')[0]
    print(result)
    print(visit(result, m).define())
    print((m.x != m.y).define())
    assert visit(result, m).define() == (m.x != m.y).define()


def test_condition_parser():
    assert m
    assert m.x
    assert m.x.define() == {'$m': 'x'}
    assert m.x > m.y
    assert (m.x > m.y).define() == {'$gt': {'x': {'$m': 'y'}}}
    assert m.x == m.y
    assert (m.x == m.y).define() == {'x': {'$m': 'y'}}
    assert m.x < m.y
    assert (m.x < m.y).define() == {'$lt': {'x': {'$m': 'y'}}}

    result = condition_parser.parseString('text')[0]
    print(result)
    print(visit_condition(result, m).define())

    result = condition_parser.parseString('""')[0]
    print(result)
    print(visit_condition(result, m))

    result = condition_parser.parseString('text != ""')[0]
    print(result)
    print(visit_condition(result, m).define())
    print((m.text != "").define())
    assert visit_condition(result, m).define() == (m.text != "").define()

    result = condition_parser.parseString('x != y')[0]
    print(result)
    print(visit_condition(result, m).define())
    print((m.x != m.y).define())
    assert visit_condition(result, m).define() == (m.x != m.y).define()

    result = condition_parser.parseString('payload.text != ""')[0]
    print(result)
    print(visit_condition(result, m).define())
    print((m.x != m.y).define())
    assert visit_condition(result, m).define() == (m.payload.text != "").define()



import ftl_events.rule_types as rule_types


def parse_rule_sets(rule_sets):
    rule_set_list = []
    for rule_set in rule_sets:
        rule_set_list.append(rule_types.RuleSet(name=rule_set['name'],
                                                hosts=rule_set['hosts'],
                                                sources=parse_event_sources(rule_set['sources']),
                                                rules=parse_rules(rule_set['rules'])))
    return rule_set_list


def parse_event_sources(sources):
    source_list = []
    for source in sources:
        name = source['name']
        del source['name']
        transform = source.get('transform')
        if transform in source:
            del source['transform']
        source_name = list(source.keys())[0]
        source_args = {k: v for k, v in source[source_name].items()}
        source_list.append(rule_types.EventSource(name=name,
                                                  source_name=source_name,
                                                  source_args=source_args,
                                                  transform=transform))

    return source_list


def parse_rules(rules):
    rule_list = []
    for rule in rules:
        name = rule.get('name')
        rule_list.append(rule_types.Rule(name=name,
                                         condition=parse_condition(rule['condition']),
                                         action=parse_action(rule['action'])))

    return rule_list


def parse_action(action):
    module_name = list(action.keys())[0]
    module_args = {k: v for k, v in action[module_name].items()}
    return rule_types.Action(module=module_name, module_args=module_args)


def parse_condition(condition):
    return rule_types.Condition(condition)



from durable.lang import *
import redis
import time
import datetime
import traceback
import sys

def unix_time(dt):
    epoch = datetime.datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return int(delta.total_seconds() * 1000.0)

def provide_durability(host, redis_host_name = 'localhost', port = 6379):
    r = redis.StrictRedis(redis_host_name, port, charset="utf-8", decode_responses=True)

    def get_hset_name(ruleset, sid):
        return 'h!{0}!{1}'.format(ruleset, sid)

    def get_list_name(ruleset, sid):
        return 'l!{0}!{1}'.format(ruleset, sid)

    def get_sset_name(ruleset):
        return 's!{0}'.format(ruleset)

    def format_message(action_type, content):
        return '{0},{1}'.format(action_type, content)

    def format_messages(results):
        if not results:
            return '[]'

        messages = ['[']
        for i in range(0, len(results)):
            messages.append(results[i])
            if i < (len(results) - 1):
                messages.append(',')

        messages.append(']') 
        return ''.join(messages)

    def store_message_callback(ruleset, sid, mid, action_type, content):
        try:
            r.hset(get_hset_name(ruleset, sid), mid, format_message(action_type, content))
        except BaseException as e:
            print(e)
            return 601

        return 0

    def delete_message_callback(ruleset, sid, mid):
        try:
            r.hdel(get_hset_name(ruleset, sid), mid)
        except BaseException as e:
            print(e)
            return 602

        return 0

    def queue_message_callback(ruleset, sid, action_type, content):
        try:
            result = r.zscore(get_sset_name(ruleset), sid)
            if not result:
                r.zadd(get_sset_name(ruleset), {sid: unix_time(datetime.datetime.now())})

            r.rpush(get_list_name(ruleset, sid), format_message(action_type, content))
        except BaseException as e:
            print(e)
            return 603 

        return 0

    def get_queued_messages_callback(ruleset, sid):
        try:
            r.zadd(get_sset_name(ruleset), {sid: unix_time(datetime.datetime.now()) + 5000})
            messages = r.lrange(get_list_name(ruleset, sid), 0, -1)
            if len(messages):
                r.delete(get_list_name(ruleset, sid))
                host.complete_get_queued_messages(ruleset, sid, format_messages(messages))
        except BaseException as e:
            print(e)
            return 604

        return 0

    def get_idle_state_callback(ruleset):
        try:
            results = r.zrangebyscore(get_sset_name(ruleset), 0, unix_time(datetime.datetime.now()))
            if (results and len(results)):
                sid = results[0]
                messages = r.hvals(get_hset_name(ruleset, sid))
                host.complete_get_idle_state(ruleset, sid, format_messages(messages))
        except BaseException as e:
            print(e)
            return 606

        return 0
            

    host.set_store_message_callback(store_message_callback)
    host.set_delete_message_callback(delete_message_callback)
    host.set_queue_message_callback(queue_message_callback)
    host.set_get_idle_state_callback(get_idle_state_callback)
    host.set_get_queued_messages_callback(get_queued_messages_callback)






from durable.lang import ruleset, rule, m
import asyncio
import jinja2
from faster_than_light import run_module, load_inventory
from ftl_events.condition_parser import parse_condition
from ftl_events.condition_types import Identifier, String, OperatorExpression


def substitute_variables(value, context):
    return jinja2.Template(value, undefined=jinja2.StrictUndefined).render(context)


def call_module(module, module_args, variables, c):
    try:
        print(c)
        variables_copy = variables.copy()
        variables_copy['event'] = str(c.m._d)
        print('running')
        asyncio.run(run_module(load_inventory('inventory.yml'),
                               ['modules'],
                               module,
                               modules=[module],
                               module_args={k: substitute_variables(v, variables_copy) for k, v in module_args.items()}))
        print('ran')
    except Exception as e:
        print(e)


def visit_condition(parsed_condition, condition):
    if isinstance(parsed_condition, Identifier):
        return condition.__getattr__(parsed_condition.value)
    if isinstance(parsed_condition, String):
        return parsed_condition.value
    if isinstance(parsed_condition, OperatorExpression):
        if parsed_condition.operator == "!=":
            return visit_condition(parsed_condition.left, condition).__ne__(visit_condition(parsed_condition.right, condition))
    if isinstance(parsed_condition, OperatorExpression):
        if parsed_condition.operator == "==":
            return visit_condition(parsed_condition.left, condition).__eq__(visit_condition(parsed_condition.right, condition))


def generate_condition(ftl_condition):
    return visit_condition(parse_condition(ftl_condition.value), m)


def make_fn(ftl_rule, variables):
    def fn(c):
        print(f'calling {ftl_rule.name}')
        call_module(ftl_rule.action.module,
                    ftl_rule.action.module_args,
                    variables,
                    c)
    return fn

def generate_rulesets(ftl_rulesets, variables):

    rulesets = []

    for ftl_ruleset in ftl_rulesets:
        a_ruleset = ruleset(ftl_ruleset.name)
        with a_ruleset:

            for ftl_rule in ftl_ruleset.rules:
                fn = make_fn(ftl_rule, variables)
                r = rule('all', True, generate_condition(ftl_rule.condition))(fn)
                print(r.define())
        rulesets.append(a_ruleset)

    return rulesets




from durable.lang import m
from pyparsing import pyparsing_common, infix_notation, OpAssoc, one_of, ParserElement, QuotedString, ZeroOrMore, Combine
ParserElement.enable_packrat()
from ftl_events.condition_types import Identifier, String, OperatorExpression



integer = pyparsing_common.signed_integer
ident = pyparsing_common.identifier
varname = Combine(ident + ZeroOrMore('.' + ident)).copy().add_parse_action(lambda toks: Identifier(toks[0]))


string1 = QuotedString("'").copy().add_parse_action(lambda toks: String(toks[0]))
string2 = QuotedString('"').copy().add_parse_action(lambda toks: String(toks[0]))

condition = infix_notation(integer | varname | string1 | string2,
                            [
                                ('!', 1, OpAssoc.RIGHT),
                                (one_of('* /'), 2, OpAssoc.LEFT),
                                (one_of('+ -'), 2, OpAssoc.LEFT),
                                (one_of('< >'), 2, OpAssoc.LEFT),
                                ('!=', 2, OpAssoc.LEFT, lambda toks: OperatorExpression(*toks[0])),
                                ('==', 2, OpAssoc.LEFT, lambda toks: OperatorExpression(*toks[0])),
                                ('>=', 2, OpAssoc.LEFT),
                                ('<=', 2, OpAssoc.LEFT),
                            ])


def parse_condition(condition_string):
    return condition.parseString(condition_string)[0]





from durable.lang import *
import durable.lang

def test_m():
    assert m
    assert m.x
    assert m.x.define() == {'$m': 'x'}
    assert m.x > m.y
    assert (m.x > m.y).define() == {'$gt': {'x': {'$m': 'y'}}}
    assert m.x == m.y
    assert (m.x == m.y).define() ==  {'x': {'$m': 'y'}}
    assert m.x < m.y
    assert (m.x < m.y).define() ==  {'$lt': {'x': {'$m': 'y'}}}


def test_ruleset():

    some_rules = ruleset('test_rules')

    assert some_rules
    assert durable.lang._rulesets
    assert durable.lang._rulesets['test_rules'] == some_rules

    assert len(durable.lang._ruleset_stack) == 0

    with some_rules:
        assert durable.lang._ruleset_stack[-1] == some_rules

    assert len(durable.lang._ruleset_stack) == 0

    assert some_rules.define() == ('test_rules', {})


def test_rules():

    some_rules = ruleset('test_rules1')

    assert some_rules
    assert durable.lang._rulesets
    assert durable.lang._rulesets['test_rules1'] == some_rules

    assert len(durable.lang._ruleset_stack) == 0

    with some_rules:
        assert durable.lang._ruleset_stack[-1] == some_rules

        def x(c):
            print('c')

        #when_all(m.x == 5)(x)
        rule('all', True, m.x == 5)(x)

    assert len(durable.lang._ruleset_stack) == 0

    assert some_rules.define() == ('test_rules1', {'r_0': {'all': [{'m': {'x': 5}}],
                                                           'run': x}})
    post('test_rules1', {'x': '5'})



def test_assert_facts():

    some_rules = ruleset('test_assert_facts')

    with some_rules:
        @when_all(+m.subject.x)
        def output(c):
            print('Fact: {0} {1} {2}'.format(c.m.subject.x, c.m.predicate, c.m.object))

    assert_fact('test_assert_facts',  { 'subject': {'x': 'Kermit'}, 'predicate': 'eats', 'object': 'flies' })

